Require Import ssreflect ssrbool ssrnat ssrfun eqtype seq path rels.
Require Import prelude prefix xfind heaps terms cancel.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* Wrap over cancelR to simplify equations of the form
      dyn d1 = dyn d2       (1)
   into
          d1 = d2           (2)
   if d1 and d2 has both the same type A.

   The idea is simple: search in the output of cancelR expressions like (1)
   and output the equation (2). The rest of the propositions are outputted
   as they are. The output of cancelR has the shape
      p1 /\ ... /\ pn
   so we only care about the /\.

   The final automated lemma, cancelRR, use another nice pattern to trigger
   the canonical structure inference mechanism. It works by equating a
   proposition p with itself, so we can have the output of cancelR in one
   side, the projector of the structure on the other side, and make them
   match with a singleton object that just match them.
*)

Structure tagged_prop := Tag {puntag :> Prop}.

Definition default_tag := Tag.
Definition dyneq_tag := default_tag.
Canonical Structure and_tag p := dyneq_tag p.

Structure form (p : Prop) := 
  Form {prop_of :> tagged_prop;
        _ : p <-> puntag prop_of}.

Program
Canonical Structure
  conj_struct p1 p2 (f1 : form p1) (f2 : form p2) := 
  @Form (p1 /\ p2) (and_tag (f1 /\ f2)) _.
Next Obligation.
by split; case: f1 f2=>[[f1]] H1 [[f2]] H2 /=; rewrite H1 H2.
Qed.

Program
Canonical Structure
  dyneq_struct A (v1 v2 : A) := 
  @Form (v1 = v2) (dyneq_tag (dyn v1 = dyn v2)) _.
Next Obligation.
by split=>[-> //|]; case/dyn_inj.
Qed.

Program
Canonical Structure
  default_struct p := 
   @Form p (default_tag p) _.
Next Obligation.
by [].
Qed.

(* pattern to trigger unification *)
Structure equate (p q : Prop) := 
  Equate {_ : p <-> q}.

Program
Definition singleton p := @Equate p p _.
Next Obligation.
by [].
Qed.

Lemma cancelRR j k t1 t2 (f1 : ast empc j t1) (f2 : ast j k t2) p (g : form p) :
        def (untag (heap_of f1)) ->
        untag (heap_of f1) = untag (heap_of f2) -> 
        equate (eval k (cancel k t1 t2)) (puntag (prop_of g)) -> p.
Proof.
case: g=>[[g]] H1 H2 H3.
simpl.
rewrite H1.
simpl.
case.
move=>H.
rewrite -H.
apply: cancelR H2 H3.
Qed.


Example ex3 h1 h2 h3 h4 x1 x2 (d1 d2 d3 d4 : nat) : 
     def ((h3 :+ (x1 :-> d1)) :+ (h1 :+ empty) :+ (x2 :-> d2)) ->
     (h3 :+ (x1 :-> d1)) :+ (h1 :+ empty) :+ 
     (x2 :-> d2) = (x2 :-> d3) :+ (h2 :+ empty :+ h3) :+ h4 :+ (x1 :-> d4) ->
     d1 = d4 /\ d2 = d3 /\ h1 = h2 :+ h4.
Proof.
move=>D H.
move: (cancelRR D H (singleton _))=>/=.
by move=>[-> [-> ->]].
Time Qed.
