Require Import ssreflect ssrfun ssrbool ssrnat seq eqtype.
Require Import heaps.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(*****************************************************************)
(* indom :                                                       *)
(*    lemma automated with Canonical Structures to prove/rewrite *)
(*    expressions with the form                                  *)
(*      x \in dom (... :+ x :-> v :+ ... )                       *)
(*    for some v. Usage:                                         *)
(*      rewrite/apply: (indom D)                                 *)
(*    where D : def (... :+ x :-> v :+ ...)                      *)
(*****************************************************************)

(* Tagging for controling the instance search *)
Structure tagged_heap := Tag {untag :> heap}.

Definition right_tag := Tag.
Definition left_tag := right_tag.
Canonical Structure found_tag h := left_tag h.

Definition invariant x (h : tagged_heap) :=
  def (untag h) -> x \in dom (untag h).

(* Main structure and instances *)
Structure find (x : ptr) := 
  Form { heap_of :> tagged_heap; 
         _ : invariant x heap_of }.

Lemma found_pf A x (v : A) : invariant x (found_tag (x :-> v)).
Proof. by rewrite /invariant defPt domPt inE /= eq_refl. Qed.

Canonical Structure ptr_found A x (v : A) :=
  @Form x (found_tag (x :-> v)) (@found_pf A x v).

Lemma left_pf x (h : heap) (f : find x) : 
        invariant x (left_tag (untag (heap_of f) :+ h)).
Proof.
case:f=>[[i]]; rewrite /invariant /= => H D.
by rewrite domUn !inE /= D (H (defUnl D)).
Qed.

Canonical Structure search_left x (h : heap) (f : find x) :=
  @Form x (left_tag (untag (heap_of f) :+ h)) (@left_pf x h f).

Lemma right_pf x (h : heap) (f : find x) : 
        invariant x (right_tag (h :+ untag (heap_of f))).
Proof.
case: f=>[[i]]; rewrite /invariant /= => H D.
by rewrite domUn !inE /= D (H (defUnr D)) orbT.
Qed.

Canonical Structure search_right x (h : heap) (f : find x) :=
  @Form x (right_tag (h :+ untag (heap_of f))) (@right_pf x h f).

(* Main lemma *)
Lemma indom (x : ptr) (f : find x) : def f -> x \in dom f.
Proof. by case: f=>[[i]]; apply. Qed.

Implicit Arguments indom [f].


(*************************************************)
(*                   Examples                    *)
(*************************************************)

(* simple example *)
Example ex1 A (x1 x2 : ptr) (v1 v2 : A) (h1 h2 : heap) :
          def (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) -> 
          if x2 \in dom (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) 
            then 1 == 1 
            else 1 == 0.
Proof.
move=>D.
by rewrite indom.
Qed.

(* same example, automatically unfolding a definition *)
Example ex2 A (x1 x2 : ptr) (v1 v2 : A) (h1 h2 : heap) :
          def (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) -> 
          if x2 \in dom (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) 
            then 1 == 1 
            else 1 == 0.
Proof.
set H := _ :+ _ :+ (_ :+ _).
move=>D.
by rewrite indom.
Qed.

Example ex4 A (x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 : ptr)
              (v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 : A) 
              (h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 : heap) :
          def (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10) -> 
         x7 \in dom (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10).
Proof.
move=>D.
Time by rewrite indom.
Qed.

Example ex5 A (x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 : ptr)
              (x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 : ptr)
              (x21 x22 x23 x24 x25 x26 x27 x28 x29 x30 : ptr)
              (v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 : A) 
              (h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 : heap) :
          def (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x16 :-> v6 :+ x17 :-> v7 :+ x18 :-> v8 :+ x19 :-> v9 :+ x20 :-> v10 :+ 
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               x26 :-> v6 :+ x27 :-> v7 :+ x28 :-> v8 :+ x29 :-> v9 :+ x30 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10) -> 
         x7 \in dom (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x16 :-> v6 :+ x17 :-> v7 :+ x18 :-> v8 :+ x19 :-> v9 :+ x20 :-> v10 :+ 
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               x26 :-> v6 :+ x27 :-> v7 :+ x28 :-> v8 :+ x29 :-> v9 :+ x30 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10).
Proof.
move=>D.
Time by rewrite indom. (* 177s! *)
Time Qed. (* 0s *)
